import java.io.*;
import java.util.*;
import java.lang.reflect.*;

public class Expression {

	static final int OP_PUSHC = 0, OP_PUSHX = 1, OP_PUSHY = 2, OP_PUSHZ = 3,
			OP_ADD = 4, OP_SUB = 5, OP_MUL = 6, OP_DIV = 7, OP_NEG = 8,
			OP_CMPEQ = 9, OP_CMPNE = 10, OP_CMPLT = 11, OP_CMPGT = 12,
			OP_CMPLE = 13, OP_CMPGE = 14, OP_GOTO = 15, OP_INVOKE = 16;

	private static final String[] opname = { "pushc", "pushx", "pushy",
			"pushz", "add", "sub", "mul", "div", "neg", "cmpeq", "cmpne",
			"cmplt", "cmpgt", "cmple", "cmpge", "goto", "invoke" };

	private int[] code;
	private double[] constPool;
	private Method[] methodPool;
	private int[] methodArg;
	private double[] stack;

	Expression(int[] code, double[] constPool, Method[] methodPool,
			int[] methodArg) {
		this.code = code;
		this.constPool = constPool;
		this.methodPool = methodPool;
		this.methodArg = methodArg;
		this.stack = new double[code.length];
	}

	private static ExpressionCompiler compiler = null;

	public Expression(String string) throws SyntaxException {
		if (compiler == null) {
			compiler = new ExpressionCompiler();
			compiler.include(Math.class);
		}
		Expression expr = compiler.compile(string);
		this.code = expr.code;
		this.constPool = expr.constPool;
		this.methodPool = expr.methodPool;
		this.methodArg = expr.methodArg;
		this.stack = expr.stack;
	}

	public double compute(double val1, double val2, double val3) {
		int sp = 0, pc, inst, addr;
		double d;

		for (pc = 0; pc < code.length;) {
			addr = code[pc++];
			inst = addr >>> 24;
			addr &= 0x00ffffff;
			switch (inst) {
			case OP_PUSHC:
				stack[sp++] = constPool[addr];
				break;
			case OP_PUSHX:
				stack[sp++] = val1;
				break;
			case OP_PUSHY:
				stack[sp++] = val2;
				break;
			case OP_PUSHZ:
				stack[sp++] = val3;
				break;
			case OP_ADD:
				d = stack[--sp];
				stack[sp - 1] += d;
				break;
			case OP_SUB:
				d = stack[--sp];
				stack[sp - 1] -= d;
				break;
			case OP_MUL:
				d = stack[--sp];
				stack[sp - 1] *= d;
				break;
			case OP_DIV:
				d = stack[--sp];
				stack[sp - 1] /= d;
				break;
			case OP_NEG:
				stack[sp - 1] = -stack[sp - 1];
				break;
			case OP_CMPEQ:
				d = stack[--sp];
				if (stack[--sp] == d)
					pc = addr;
				break;
			case OP_CMPNE:
				d = stack[--sp];
				if (stack[--sp] != d)
					pc = addr;
				break;
			case OP_CMPLT:
				d = stack[--sp];
				if (stack[--sp] < d)
					pc = addr;
				break;
			case OP_CMPGT:
				d = stack[--sp];
				if (stack[--sp] > d)
					pc = addr;
				break;
			case OP_CMPLE:
				d = stack[--sp];
				if (stack[--sp] <= d)
					pc = addr;
				break;
			case OP_CMPGE:
				d = stack[--sp];
				if (stack[--sp] >= d)
					pc = addr;
				break;
			case OP_GOTO:
				pc = addr;
				break;
			case OP_INVOKE:
				Method method = methodPool[addr];
				int nparams = methodArg[addr];
				Double[] params = new Double[nparams];
				for (int i = nparams - 1; i >= 0; i--)
					params[i] = new Double(stack[--sp]);
				try {
					Number ret = (Number) method.invoke(null, params);
					stack[sp++] = ret.doubleValue();
				} catch (IllegalAccessException iace) {
					throw new RuntimeException("[BUG] Method#invoke raised: "
							+ iace);
				} catch (IllegalArgumentException iare) {
					throw new RuntimeException("[BUG] Method#invoke raised: "
							+ iare);
				} catch (InvocationTargetException ite) {
					throw new RuntimeException(method + " raised: ["
							+ ite.getTargetException() + "]");
				}
				break;
			default:
				throw new RuntimeException("[BUG] unknown inst-code: " + inst);
			}
		}
		if (sp != 1)
			throw new RuntimeException("[BUG] stack level mismatch");

		return stack[0];
	}

	public double compute(double val1, double val2) {
		return compute(val1, val2, 0.0);
	}

	public double compute(double val1) {
		return compute(val1, 0.0, 0.0);
	}

	public void disassemble(PrintStream out) {
		int pc, inst, addr;

		for (pc = 0; pc < code.length;) {
			out.print(pc + ": ");
			addr = code[pc++];
			inst = addr >>> 24;
			addr &= 0x00ffffff;
			if (0 <= inst && inst < opname.length) {
				out.print(opname[inst]);
				switch (inst) {
				case OP_PUSHC:
					out.print(" <" + constPool[addr] + ">");
					break;
				case OP_CMPEQ:
				case OP_CMPNE:
				case OP_CMPLT:
				case OP_CMPGT:
				case OP_CMPLE:
				case OP_CMPGE:
				case OP_GOTO:
					out.print(" " + addr);
					break;
				case OP_INVOKE:
					out.println(" <" + methodPool[addr] + ">");
					break;
				}
				out.println();
			} else {
				out.println("*** unknown inst-code: " + inst);
			}
		}
	}
}

class ExpressionCompiler {
	private Hashtable funcHash;
	private Hashtable constHash;
	private Tokenizer st;
	private Vector code;
	private Vector constPool;
	private Vector methodPool;
	private Vector methodArg;
	private String var1, var2, var3;

	public ExpressionCompiler() {
		funcHash = new Hashtable();
		constHash = new Hashtable();
		code = new Vector();
		constPool = new Vector();
		methodPool = new Vector();
		methodArg = new Vector();
	}

	public Expression compile(String text, String var1, String var2, String var3)
			throws SyntaxException {
		st = new Tokenizer(new StringReader(text));

		this.var1 = var1;
		this.var2 = var2;
		this.var3 = var3;

		code.removeAllElements();
		constPool.removeAllElements();
		methodPool.removeAllElements();
		methodArg.removeAllElements();

		nextToken();
		parseExpr();
		if (st.ttype != Tokenizer.TT_EOF)
			throw new SyntaxException("unexpected token: " + st);

		int i;
		int[] a_code = new int[code.size()];
		for (i = 0; i < code.size(); i++)
			a_code[i] = ((Integer) code.elementAt(i)).intValue();
		double[] a_constPool = new double[constPool.size()];
		for (i = 0; i < constPool.size(); i++)
			a_constPool[i] = ((Double) constPool.elementAt(i)).doubleValue();
		Method[] a_methodPool = new Method[methodPool.size()];
		for (i = 0; i < methodPool.size(); i++)
			a_methodPool[i] = (Method) methodPool.elementAt(i);
		int[] a_methodArg = new int[methodArg.size()];
		for (i = 0; i < methodArg.size(); i++)
			a_methodArg[i] = ((Integer) methodArg.elementAt(i)).intValue();
		code.removeAllElements();
		constPool.removeAllElements();
		methodPool.removeAllElements();
		methodArg.removeAllElements();
		st = null;
		var1 = var2 = var3 = null;
		return new Expression(a_code, a_constPool, a_methodPool, a_methodArg);
	}

	public Expression compile(String text, String var1, String var2)
			throws SyntaxException {
		return compile(text, var1, var2, null);
	}

	public Expression compile(String text, String var1) throws SyntaxException {
		return compile(text, var1, null, null);
	}

	public Expression compile(String text) throws SyntaxException {
		return compile(text, "x", "y", "z");
	}

	private int nextToken() {
		try {
			return st.nextToken();
		} catch (IOException ioe) {
			throw new RuntimeException("[BUG] Tokenizer#nextToken raised: "
					+ ioe);
		}
	}

	private void expectToken(int ttype) throws SyntaxException {
		if (st.ttype != ttype)
			throw new SyntaxException("'" + Tokenizer.toString(ttype)
					+ "' expected");
		nextToken();
	}

	private void parseFact() throws SyntaxException {
		if (st.ttype == Tokenizer.TT_NUMBER) {
			addCode(Expression.OP_PUSHC, st.nval);
			nextToken();
		} else if (st.ttype == Tokenizer.TT_WORD) {
			String name = st.sval;
			nextToken();
			if (st.ttype == '(') {
				int nparams = 0;
				nextToken();
				if (st.ttype != ')') {
					while (true) {
						parseExpr();
						nparams++;
						if (st.ttype == ')')
							break;
						expectToken(',');
					}
				}
				nextToken();
				Method m = findFunction(name, nparams);
				if (m == null)
					throw new SyntaxException("no such function: " + name);
				addCode(Expression.OP_INVOKE, m, nparams);
			} else {
				if (name.equals(var1))
					addCode(Expression.OP_PUSHX);
				else if (name.equals(var2))
					addCode(Expression.OP_PUSHY);
				else if (name.equals(var3))
					addCode(Expression.OP_PUSHZ);
				else {
					Double d = findConstant(name);
					if (d == null)
						throw new SyntaxException(
								"no such variable or constant: " + name);
					addCode(Expression.OP_PUSHC, d.doubleValue());
				}
			}
		} else if (st.ttype == '(') {
			nextToken();
			parseExpr();
			expectToken(')');
		} else if (st.ttype == '-') {
			nextToken();
			parseFact();
			int size = code.size();
			int inst, addr;
			addr = ((Integer) code.elementAt(size - 1)).intValue();
			inst = addr >> 24;
			addr &= 0x00ffffff;
			if (inst == Expression.OP_PUSHC) {
				double d = ((Double) constPool.elementAt(addr)).doubleValue();
				constPool.setElementAt(new Double(-d), addr);
			} else {
				addCode(Expression.OP_NEG);
			}
		} else if (st.ttype == Tokenizer.TT_IF) {
			nextToken();
			expectToken('(');
			LabelRef l1 = new LabelRef();
			parseConditional(false, l1);
			expectToken(')');
			parseExpr();
			expectToken(Tokenizer.TT_ELSE);
			LabelRef l2 = new LabelRef();

			addCode(Expression.OP_GOTO, 0);
			l2.addReference(code.size() - 1);
			l1.resolve(code, code.size());
			parseExpr();
			l2.resolve(code, code.size());
		} else {
			throw new SyntaxException("parse error at token: " + st);
		}
	}

	private void parseTerm() throws SyntaxException {
		parseFact();
		while (st.ttype == '*' || st.ttype == '/') {
			int op = st.ttype;
			nextToken();
			parseFact();
			addCode(op == '*' ? Expression.OP_MUL : Expression.OP_DIV);
		}
	}

	private void parseExpr() throws SyntaxException {
		parseTerm();
		while (st.ttype == '+' || st.ttype == '-') {
			int op = st.ttype;
			nextToken();
			parseTerm();
			addCode(op == '+' ? Expression.OP_ADD : Expression.OP_SUB);
		}
	}

	private void parseCompare(boolean negative, LabelRef falseLabel)
			throws SyntaxException {
		parseExpr();
		if (st.ttype == Tokenizer.TT_EQ || st.ttype == Tokenizer.TT_NEQ
				|| st.ttype == '<' || st.ttype == '>'
				|| st.ttype == Tokenizer.TT_LE || st.ttype == Tokenizer.TT_GE) {
			int op = st.ttype;
			nextToken();
			parseExpr();
			if (negative) {
				addCode(
						op == Tokenizer.TT_EQ ? Expression.OP_CMPEQ
								: op == Tokenizer.TT_NEQ ? Expression.OP_CMPNE
										: op == '<' ? Expression.OP_CMPLT
												: op == '>' ? Expression.OP_CMPGT
														: op == Tokenizer.TT_LE ? Expression.OP_CMPLE
																: Expression.OP_CMPGE,
						0);
			} else {
				addCode(
						op == Tokenizer.TT_EQ ? Expression.OP_CMPNE
								: op == Tokenizer.TT_NEQ ? Expression.OP_CMPEQ
										: op == '<' ? Expression.OP_CMPGE
												: op == '>' ? Expression.OP_CMPLE
														: op == Tokenizer.TT_LE ? Expression.OP_CMPGT
																: Expression.OP_CMPLT,
						0);
			}
			falseLabel.addReference(code.size() - 1);
		} else
			throw new SyntaxException("comparison operator expected");
	}

	private void parseLogical(boolean negative, LabelRef falseLabel)
			throws SyntaxException {
		if (st.ttype == '(') {
			nextToken();
			parseConditional(negative, falseLabel);
			expectToken(')');
		} else if (st.ttype == '!') {
			nextToken();
			parseLogical(!negative, falseLabel);
		} else {
			parseCompare(negative, falseLabel);
		}
	}

	private void parseAnd(boolean negative, LabelRef falseLabel)
			throws SyntaxException {
		if (!negative) {
			parseLogical(false, falseLabel);
			while (st.ttype == Tokenizer.TT_AND2) {
				nextToken();
				parseLogical(false, falseLabel);
			}
		} else {
			LabelRef l1 = new LabelRef();
			parseLogical(false, l1);
			while (st.ttype == Tokenizer.TT_AND2) {
				nextToken();
				parseLogical(false, l1);
			}
			addCode(Expression.OP_GOTO, 0);
			falseLabel.addReference(code.size() - 1);
			l1.resolve(code, code.size());
		}
	}

	private void parseOr(boolean negative, LabelRef falseLabel)
			throws SyntaxException {
		if (!negative) {
			LabelRef l1 = new LabelRef();
			parseAnd(true, l1);
			while (st.ttype == Tokenizer.TT_OR2) {
				nextToken();
				parseAnd(true, l1);
			}
			addCode(Expression.OP_GOTO, 0);
			falseLabel.addReference(code.size() - 1);
			l1.resolve(code, code.size());
		} else {
			parseAnd(true, falseLabel);
			while (st.ttype == Tokenizer.TT_OR2) {
				nextToken();
				parseAnd(true, falseLabel);
			}
		}
	}

	private void parseConditional(boolean negative, LabelRef falseLabel)
			throws SyntaxException {
		parseOr(negative, falseLabel);
	}

	private void addCode(int inst, int addr) throws SyntaxException {
		if (addr < 0 || 0x00ffffff < addr)
			throw new SyntaxException(
					"expression too complex! (address overflow)");
		code.addElement(new Integer((inst << 24) | addr));
	}

	private void addCode(int inst) throws SyntaxException {
		addCode(inst, 0);
	}

	private void addCode(int inst, double value) throws SyntaxException {
		addCode(inst, constPool.size());
		constPool.addElement(new Double(value));
	}

	private void addCode(int inst, Method method, int nparams)
			throws SyntaxException {
		addCode(inst, methodPool.size());
		methodPool.addElement(method);
		methodArg.addElement(new Integer(nparams));
	}

	private Method findFunction(String name, int nparams) {
		return (Method) funcHash.get(name + "#" + nparams);
	}

	private Double findConstant(String name) {
		Field f = (Field) constHash.get(name);
		if (f == null)
			return null;
		try {
			return ((Double) f.get(null));
		} catch (IllegalAccessException iae) {
			throw new RuntimeException("[BUG] Field#get raised: " + iae);
		}
	}

	public void include(Class clazz) {
		if (!Modifier.isPublic(clazz.getModifiers()))
			throw new IllegalArgumentException("Class is not public: " + clazz);

		Method[] methods = clazz.getMethods();
		for (int i = 0; i < methods.length; i++) {
			Method method = methods[i];
			if (methodIsUsableAsFunction(method)) {
				funcHash.put(method.getName() + "#"
						+ method.getParameterTypes().length, method);
			}
		}

		Field[] fields = clazz.getFields();
		for (int i = 0; i < fields.length; i++) {
			Field field = fields[i];
			if (fieldIsUsableAsConstant(field)) {
				constHash.put(field.getName(), field);
			}
		}
	}

	private static boolean methodIsUsableAsFunction(Method method) {
		int mod = method.getModifiers();
		Class type = method.getReturnType();

		if (Modifier.isPublic(mod)
				&& Modifier.isStatic(mod)
				&& !Modifier.isAbstract(mod)
				&& (type == byte.class || type == short.class
						|| type == int.class || type == long.class
						|| type == float.class || type == double.class)) {
			Class[] paramtypes = method.getParameterTypes();
			int i;
			for (i = 0; i < paramtypes.length; i++) {
				if (paramtypes[i] != double.class)
					break;
			}
			if (i >= paramtypes.length
					&& method.getExceptionTypes().length == 0) {
				return true;
			}
		}
		return false;
	}

	private static boolean fieldIsUsableAsConstant(Field field) {
		int mod = field.getModifiers();
		Class type = field.getType();

		if (Modifier.isPublic(mod)
				&& Modifier.isStatic(mod)
				&& !Modifier.isAbstract(mod)
				&& Modifier.isFinal(mod)
				&& (type == byte.class || type == short.class
						|| type == int.class || type == long.class
						|| type == float.class || type == double.class)) {
			return true;
		}
		return false;
	}

	public Method[] getRegisteredMethods() {
		int size = funcHash.size();
		Method[] res = new Method[size];
		Enumeration e = funcHash.elements();

		for (int i = 0; i < size; i++)
			res[i] = (Method) e.nextElement();

		return res;
	}

	public Field[] getRegisteredFields() {
		int size = constHash.size();
		Field[] res = new Field[size];
		Enumeration e = constHash.elements();

		for (int i = 0; i < size; i++)
			res[i] = (Field) e.nextElement();

		return res;
	}
}

class LabelRef {
	private int[] list;
	private int count;

	public LabelRef() {
		list = new int[4];
		count = 0;
	}

	public void addReference(int addr) {
		if (count == list.length) {
			int[] list2 = new int[count * 2];
			System.arraycopy(list, 0, list2, 0, count);
			list = list2;
		}
		list[count++] = addr;
	}

	public void resolve(Vector code, int addr) throws SyntaxException {
		if (addr < 0 || 0x00ffffff < addr)
			throw new SyntaxException(
					"expression too complex! (address overflow)");
		for (int i = 0; i < count; i++) {
			int bits = ((Integer) code.elementAt(list[i])).intValue();
			bits = (bits & 0xff000000) | addr;
			code.setElementAt(new Integer(bits), list[i]);
		}
	}
}

class Tokenizer {
	public static final int TT_EOF = -1, TT_NUMBER = -2, TT_WORD = -3,
			TT_LE = -4, TT_GE = -5, TT_EQ = -6, TT_NEQ = -7, TT_OR2 = -8,
			TT_AND2 = -9, TT_IF = -10, TT_ELSE = -11, TT_NOTHING = -9999;

	public int ttype;
	public double nval;
	public String sval;

	private Reader input;
	private int peekc;
	private StringBuffer sbuf = new StringBuffer();

	public Tokenizer(Reader in) {
		this.input = in;
		this.ttype = TT_NOTHING;
	}

	public int nextToken() throws IOException {
		int c;

		sval = null;
		if (ttype == TT_NOTHING) {
			peekc = input.read();
		}
		c = peekc;

		if (c == -1)
			return (ttype = TT_EOF);
		while (Character.isWhitespace((char) c)) {
			if ((c = input.read()) == -1)
				return (ttype = TT_EOF);
		}

		if (('0' <= c && c <= '9') || c == '.') {
			boolean dot = false;
			double val = 0, div = 1;
			while (true) {
				if ('0' <= c && c <= '9') {
					val = val * 10 + c - '0';
					if (dot)
						div *= 10;
				} else if (c == '.' && !dot) {
					dot = true;
				} else
					break;
				c = input.read();
			}
			peekc = c;
			if (dot)
				val /= div;
			nval = val;
			return (ttype = TT_NUMBER);
		} else if (Character.isJavaIdentifierStart((char) c)) {
			sbuf.setLength(0);
			sbuf.append((char) c);
			while ((c = input.read()) != -1
					&& Character.isJavaIdentifierPart((char) c))
				sbuf.append((char) c);
			peekc = c;
			sval = sbuf.toString();
			if (sval.equals("if"))
				return (ttype = TT_IF);
			else if (sval.equals("else"))
				return (ttype = TT_ELSE);
			return (ttype = TT_WORD);
		} else if (c == '<' || c == '>' || c == '=' || c == '!') {
			if ((peekc = input.read()) == '=') {
				peekc = input.read();
				return (ttype = (c == '<' ? TT_LE : c == '>' ? TT_GE
						: c == '=' ? TT_EQ : TT_NEQ));
			}
			return (ttype = c);
		} else if (c == '&' || c == '|') {
			if ((peekc = input.read()) == c) {
				peekc = input.read();
				return (ttype = (c == '&' ? TT_AND2 : TT_OR2));
			}
			return (ttype = c);
		} else {
			peekc = input.read();
			return (ttype = c);
		}
	}

	public String toString() {
		String r;
		switch (ttype) {
		case TT_NUMBER:
			r = String.valueOf(nval);
			break;
		case TT_WORD:
			r = sval;
			break;
		default:
			r = toString(ttype);
			break;
		}
		return r;
	}

	public static String toString(int ttype) {
		String r;

		switch (ttype) {
		case TT_EOF:
			r = "EOF";
			break;
		case TT_NUMBER:
			r = "number";
			break;
		case TT_WORD:
			r = "identifier";
			break;
		case TT_LE:
			r = "'<='";
			break;
		case TT_GE:
			r = "'>='";
			break;
		case TT_EQ:
			r = "'=='";
			break;
		case TT_NEQ:
			r = "'!='";
			break;
		case TT_OR2:
			r = "'||'";
			break;
		case TT_AND2:
			r = "'&&'";
			break;
		case TT_IF:
			r = "'if'";
			break;
		case TT_ELSE:
			r = "'else'";
			break;
		case TT_NOTHING:
			r = "NOTHING";
			break;
		default:
			if (ttype < 0)
				r = "TT_???(" + ttype + ")";
			else
				r = "'" + (char) ttype + "'";
			break;
		}
		return r;
	}
}

class SyntaxException extends Exception {

	public SyntaxException() {
		super();
	}

	public SyntaxException(String s) {
		super(s);
	}
}
