import java.util.*;

import JSci.maths.statistics.NormalDistribution;

public abstract class Genome<T> {

	public T chromosome;

	public GeneticAlgorithm<T> ga;

	public int size;

	private int fitnessRank, deviationRank;

	protected double fitnessValue;

	protected double deviation;

	protected boolean isEvaluated, randomInitial = true;

	public abstract void mutate(double rate);

	public abstract void crossOver(Genome<T> g, double rate);

	public abstract double getFitnessValue();

	public abstract void setDeviation(ArrayList<Genome<T>> g);

	public abstract Genome<T> clone();

	public abstract Genome<T> randomClone();

	public abstract String printChromosome();

	public abstract boolean equals(Object o);

	public int getFitnessRank() {
		return this.fitnessRank;
	}

	public double getDeviation() {
		return this.deviation;
	}

	public int getDeviationRank() {
		return this.deviationRank;
	}

	public void setFitnessRank(int fitnessRank) {
		this.fitnessRank = fitnessRank;
	}

	public void setDeviationRank(int deviationRank) {
		this.deviationRank = deviationRank;
	}

	public int findOneRandomPoint() {
		return (int) (ga.random.nextDouble() * size);
	}
}

class GenomeComparator implements Comparator<Genome<?>> {

	public static int FITNESS = 0, DEVIATION = 1, COMBINE = 2;

	private boolean maximize;

	private int sortType;

	public GenomeComparator(boolean maximize) {
		this(maximize, 0);
	}

	public GenomeComparator(boolean maximize, int sortType) {
		super();
		this.maximize = maximize;
		this.sortType = sortType;
	}

	public int compare(Genome<?> g1, Genome<?> g2) {
		if (sortType == FITNESS) {
			double g1Fitness = g1.getFitnessValue(), g2Fitness = g2
					.getFitnessValue();
			if (g1Fitness < g2Fitness)
				if (maximize)
					return 1;
				else
					return -1;
			else if (g1Fitness > g2Fitness)
				if (maximize)
					return -1;
				else
					return 1;
		} else if (sortType == DEVIATION) {
			if (g1.getDeviation() > g2.getDeviation())
				return -1;
			else if (g1.getDeviation() < g2.getDeviation())
				return 1;
		} else if (sortType == COMBINE) {
			if (g1.getFitnessRank() + g1.getDeviationRank() < g2
					.getFitnessRank()
					+ g2.getDeviationRank())
				return -1;
			else if (g1.getFitnessRank() + g1.getDeviationRank() == g2
					.getFitnessRank()
					+ g2.getDeviationRank()) {
				if (g1.getFitnessRank() < g2.getFitnessRank())
					return -1;
				else
					return 1;
			} else
				return 1;
		}
		return 0;
	}
}

class BitStringGenome extends Genome<int[]> {

	protected BitStringTestFunction bf;

	public BitStringGenome(int size, BitStringTestFunction bf,
			GeneticAlgorithm<int[]> ga) {
		this.size = size;
		this.bf = bf;
		this.ga = ga;
		this.chromosome = new int[size];
		if (randomInitial) {
			randomChromosome();
		}
	}

	protected void randomChromosome() {
		for (int i = 0; i < chromosome.length; i++) {
			chromosome[i] = (int) Math.round(ga.random.nextDouble());
		}
	}

	public BitStringGenome(BitStringGenome g) {
		this.size = g.size;
		this.bf = g.bf;
		this.ga = g.ga;
		this.fitnessValue = g.fitnessValue;
		this.isEvaluated = g.isEvaluated;
		this.chromosome = Arrays.copyOf(g.chromosome, this.size);
	}

	public BitStringGenome clone() {
		return new BitStringGenome(this);
	}

	public Genome<int[]> randomClone() {
		return new BitStringGenome(this.size, this.bf, this.ga);
	}

	@SuppressWarnings("unchecked")
	public boolean equals(Object o) {
		Genome<int[]> g;
		if (o.getClass() != this.getClass())
			return false;
		else
			g = (Genome<int[]>) o;
		for (int i = 0; i < this.chromosome.length; i++) {
			if (this.chromosome[i] != g.chromosome[i])
				return false;
		}
		return true;
	}

	public void crossOver(Genome<int[]> g, double rate) {
		int crossOverPoint = findOneRandomPoint();
		if (rate < ga.random.nextDouble())
			return;
		onePointCrossOver(g, crossOverPoint);
		isEvaluated = false;
		g.isEvaluated = false;
	}

	public void onePointCrossOver(Genome<int[]> g, int crossOverPoint) {
		int[] g1 = new int[size];
		int[] g2 = new int[size];
		for (int i = 0; i < size; i++) {
			if (i < crossOverPoint)
				g1[i] = this.chromosome[i];
			else
				g2[i] = this.chromosome[i];
		}
		for (int i = 0; i < size; i++) {
			if (i < crossOverPoint)
				g2[i] = g.chromosome[i];
			else
				g1[i] = g.chromosome[i];
		}
		this.chromosome = g1;
		g.chromosome = g2;
	}

	public void mutate(double rate) {
		for (int i = 0; i < this.chromosome.length; i++) {
			if (ga.random.nextDouble() < rate) {
				this.chromosome[i] = this.chromosome[i] ^ 1;
				isEvaluated = false;
			}
		}
	}

	public double getFitnessValue() {
		if (!isEvaluated) {
			fitnessValue = bf.getFunctionValue(this.chromosome);
			isEvaluated = true;
			ga.evaluationCount++;
		}
		return fitnessValue;
	}

	public String printChromosome() {
		boolean printFitness = true;
		String s = "";
		if (printFitness)
			s += " Fitness: " + this.getFitnessValue() + " Chromosome: ";
		s += printArray(chromosome);
		return s;
	}

	public void setDeviation(ArrayList<Genome<int[]>> genomeList) {
		double sum = 0;
		for (Genome<int[]> g : genomeList)
			for (int i = 0; i < chromosome.length; i++) {
				sum += Math.sqrt(Math.abs(chromosome[i] - g.chromosome[i]));
			}
		this.deviation = sum;
	}

	public static void swap(int[] g1, int[] g2, int i, int j) {
		int temp = g1[i];
		g1[i] = g2[j];
		g2[j] = temp;
	}

	public static String printArray(int[] array) {
		String s = "";
		for (int i : array) {
			s += i + ", ";
		}
		s = s.substring(0, s.lastIndexOf(','));
		return s;
	}
}

class PermutationGenome extends BitStringGenome {

	public PermutationGenome(int size, BitStringTestFunction bf,
			PermutationGeneticAlgorithm ga) {
		super(size, bf, ga);
	}

	protected void randomChromosome() {
		for (int i = 0; i < chromosome.length; i++)
			chromosome[i] = i;
		for (int i = 0; i < size; i++) {
			int j = (int) (ga.random.nextDouble() * size);
			swap(chromosome, chromosome, i, j);
		}
	}

	public PermutationGenome(BitStringGenome g) {
		super(g);
	}

	public PermutationGenome clone() {
		return new PermutationGenome(this);
	}

	public PermutationGenome randomClone() {
		return new PermutationGenome(this.size, this.bf,
				(PermutationGeneticAlgorithm) this.ga);
	}

	public void crossOver(Genome<int[]> g, double rate) {
		int crossOverPoint;
		for (int i = 0; i < size / 3; i++)
			if (rate > ga.random.nextDouble()) {
				crossOverPoint = findOneRandomPoint();
				UPMXCrossOver(g, crossOverPoint);
				isEvaluated = false;
				g.isEvaluated = false;
			}
	}

	public void UPMXCrossOver(Genome<int[]> g, int crossOverPoint) {
		for (int i = 0; i < size; i++) {
			if (g.chromosome[i] == this.chromosome[crossOverPoint]) {
				for (int j = 0; j < size; j++) {
					if (this.chromosome[j] == g.chromosome[crossOverPoint]) {
						swap(this.chromosome, g.chromosome, crossOverPoint,
								crossOverPoint);
						swap(this.chromosome, g.chromosome, j, i);
						break;
					}
				}
				break;
			}
		}
	}

	public void mutate(double rate) {
		if (rate > 0) {
			int i = findOneRandomPoint(), j = findOneRandomPoint();
			swap(this.chromosome, this.chromosome, i, j);
			isEvaluated = false;
		}
	}
}

class GridGenome extends BitStringGenome {

	public GridGenome(int size, GridGeneticAlgorithm ga) {
		super(size, null, ga);
	}

	protected void randomChromosome() {
		GridGeneticAlgorithm gga = (GridGeneticAlgorithm) ga;
		int centerArea = (int) Math.round(gga.UPPER_BOUND
				* gga.CENTRAL_AREA_PERCENTAGE);
		int lowBox = (int) (gga.UPPER_BOUND / 2.0 - centerArea), highBox = (int) Math
				.ceil(gga.UPPER_BOUND / 2.0 + centerArea);
		if (gga.randomInitial) {
			do {
				this.chromosome[0] = (int) (ga.random.nextDouble() * gga.UPPER_BOUND);
				this.chromosome[1] = (int) (ga.random.nextDouble() * gga.UPPER_BOUND);
			} while ((this.chromosome[0] >= lowBox && this.chromosome[0] <= highBox)
					|| (this.chromosome[1] >= lowBox && this.chromosome[1] <= highBox));
		}

	}

	public GridGenome(GridGenome g) {
		super(g);
	}

	public GridGenome clone() {
		return new GridGenome(this);
	}

	public GridGenome randomClone() {
		return new GridGenome(this.size, (GridGeneticAlgorithm) this.ga);
	}

	public void crossOver(Genome<int[]> g, double rate) {
		int crossOverPoint = 1;
		if (rate < ga.random.nextDouble())
			return;
		onePointCrossOver(g, crossOverPoint);
		isEvaluated = false;
		g.isEvaluated = false;
	}

	public void mutate(double rate) {
		for (int i = 0; i < this.chromosome.length; i++) {
			if (ga.random.nextDouble() < rate) {
				if (this.chromosome[i] <= ((GridGeneticAlgorithm) ga).LOWER_BOUND)
					this.chromosome[i]++;
				else if (this.chromosome[i] >= ((GridGeneticAlgorithm) ga).UPPER_BOUND)
					this.chromosome[i]--;
				else {
					int x = (int) Math.pow(-1,
							(int) (ga.random.nextDouble() * 10) % 2);
					this.chromosome[i] += x;
				}
				isEvaluated = false;
			}
		}
	}

	public double getFitnessValue() {
		if (!isEvaluated) {
			fitnessValue = ((GridGeneticAlgorithm) ga).FITNESS[chromosome[0]][chromosome[1]];
			isEvaluated = true;
			ga.evaluationCount++;
		}
		return fitnessValue;
	}
}

class RealCodedGenome extends Genome<double[]> {

	protected double upperDomain = 1.0;

	protected double lowerDomain = 0.0;

	public static double alpha = 0.1;

	public RealCodedGenome(int size, RealCodedGeneticAlgorithm ga) {
		this.size = size;
		this.ga = ga;
		this.lowerDomain = SingleObjectiveTestFunction
				.getLowerDomain(ga.testNumber);
		this.upperDomain = SingleObjectiveTestFunction
				.getUpperDomain(ga.testNumber);
		this.chromosome = new double[size];
		if (randomInitial) {
			randomChromosome();
		}
	}

	public RealCodedGenome(RealCodedGenome g) {
		this.size = g.size;
		this.ga = g.ga;
		this.lowerDomain = g.lowerDomain;
		this.upperDomain = g.upperDomain;
		this.fitnessValue = g.fitnessValue;
		this.isEvaluated = g.isEvaluated;
		this.chromosome = Arrays.copyOf(g.chromosome, this.size);
	}

	protected void randomChromosome() {
		for (int i = 0; i < chromosome.length; i++) {
			chromosome[i] = ga.random.nextDouble()
					* (this.upperDomain - this.lowerDomain) + this.lowerDomain;
		}
	}

	public RealCodedGenome clone() {
		return new RealCodedGenome(this);
	}

	public RealCodedGenome randomClone() {
		return new RealCodedGenome(this.size,
				(RealCodedGeneticAlgorithm) this.ga);
	}

	@SuppressWarnings("unchecked")
	public boolean equals(Object o) {
		Genome<double[]> g;
		if (o.getClass() != this.getClass())
			return false;
		else
			g = (Genome<double[]>) o;
		for (int i = 0; i < this.chromosome.length; i++) {
			if (this.chromosome[i] != g.chromosome[i])
				return false;
		}
		return true;
	}

	public void crossOver(Genome<double[]> g, double rate) {
		if (rate > ga.random.nextDouble()) {
			// singlePointCrossOver(g, rate);
			blxAlphaCrossOver(g, rate);
			isEvaluated = false;
			g.isEvaluated = false;
		}
	}

	public void singlePointCrossOver(Genome<double[]> g, double rate) {
		int crossOverPoint = findOneRandomPoint();
		double[] g1 = new double[size];
		double[] g2 = new double[size];
		for (int i = 0; i < size; i++) {
			if (i < crossOverPoint)
				g1[i] = this.chromosome[i];
			else
				g2[i] = this.chromosome[i];
		}
		for (int i = 0; i < size; i++) {
			if (i < crossOverPoint)
				g2[i] = g.chromosome[i];
			else
				g1[i] = g.chromosome[i];
		}
		this.chromosome = g1;
		g.chromosome = g2;
	}

	public void blxAlphaCrossOver(Genome<double[]> g, double rate) {
		double[] g1 = new double[size];
		double[] g2 = new double[size];
		double[] d = new double[size];
		double[] min = new double[size];
		double[] max = new double[size];
		for (int i = 0; i < size; i++) {
			d[i] = Math.abs(this.chromosome[i] - g.chromosome[i]);
			min[i] = Math.min(this.chromosome[i], g.chromosome[i]);
			min[i] = Math.max(min[i] - d[i] * alpha, lowerDomain);
			max[i] = Math.max(this.chromosome[i], g.chromosome[i]);
			max[i] = Math.min(max[i] + d[i] * alpha, upperDomain);
			g1[i] = (max[i] - min[i]) * ga.random.nextDouble() + min[i];
			g2[i] = (max[i] - min[i]) * ga.random.nextDouble() + min[i];
		}
		this.chromosome = g1;
		g.chromosome = g2;
	}

	public void mutate(double rate) {
		for (int i = 0; i < this.chromosome.length; i++) {
			if (ga.random.nextDouble() < rate) {
				this.chromosome[i] = this.gaussianMutate(i);
				isEvaluated = false;
			}
		}
	}

	protected double gaussianMutate(int i) {
		double stdDev = calculateSD();
		NormalDistribution nd = new NormalDistribution();
		double mutatedValue = stdDev * nd.inverse(ga.random.nextDouble());
		double result = this.chromosome[i] + mutatedValue;
		while (result > upperDomain || result < lowerDomain) {
			mutatedValue /= 2;
			result = this.chromosome[i] + mutatedValue;
		}
		return result;
	}

	protected double calculateSD() {
		double stdDev = 0, mean = calculateMean();
		for (int i = 0; i < size; i++)
			stdDev += (chromosome[i] - mean) * (chromosome[i] - mean);
		stdDev /= size;
		stdDev = Math.sqrt(stdDev);
		return stdDev;
	}

	protected double calculateMean() {
		double sum = 0;
		for (int i = 0; i < size; i++)
			sum += chromosome[i];
		sum /= size;
		return sum;
	}

	public double getFitnessValue() {
		if (!isEvaluated) {
			if (GeneticAlgorithm.USE_PREDEFINED_FUNCTION)
				fitnessValue = SingleObjectiveTestFunction
						.calculateFunctionValue(
								((RealCodedGeneticAlgorithm) ga).testNumber,
								this.chromosome);
			else
				fitnessValue = ((RealCodedGeneticAlgorithm) ga).expr[0]
						.compute(this.chromosome[0], this.chromosome[1],
								this.chromosome[2]);
			isEvaluated = true;
			ga.evaluationCount++;
		}
		return fitnessValue;
	}

	public String printChromosome() {
		boolean printFitness = false;
		String s = "";
		for (double i : chromosome) {
			s += i + ", ";
		}
		s = s.substring(0, s.lastIndexOf(','));
		if (printFitness)
			s += " Fitness :" + this.getFitnessValue();
		return s;
	}

	public void setDeviation(ArrayList<Genome<double[]>> genomeList) {
		double sum = 0;
		for (Genome<double[]> g : genomeList)
			for (int i = 0; i < chromosome.length; i++) {
				sum += (chromosome[i] - g.chromosome[i])
						* (chromosome[i] - g.chromosome[i]);
			}
		this.deviation = sum;
	}
}

class MultiObjectiveRealCodedGenome extends RealCodedGenome {

	private int rank;

	public int objectiveNum;

	public double[] objectiveValue;

	public int getRank() {
		return rank;
	}

	public void setRank(int rank) {
		this.rank = rank;
	}

	public MultiObjectiveRealCodedGenome clone() {
		return new MultiObjectiveRealCodedGenome(this);
	}

	public MultiObjectiveRealCodedGenome randomClone() {
		return new MultiObjectiveRealCodedGenome(this.size, this.objectiveNum,
				(MultiObjectiveRealCodedGeneticAlgorithm) this.ga);
	}

	public MultiObjectiveRealCodedGenome(int size, int objectiveNum,
			MultiObjectiveRealCodedGeneticAlgorithm ga) {
		super(size, ga);
		this.lowerDomain = MultiObjectiveTestFunction
				.getLowerDomain(ga.testNumber);
		this.upperDomain = MultiObjectiveTestFunction
				.getUpperDomain(ga.testNumber);
		this.objectiveNum = objectiveNum;
		objectiveValue = new double[this.objectiveNum];
		if (randomInitial) {
			randomChromosome();
		}
	}

	public MultiObjectiveRealCodedGenome(MultiObjectiveRealCodedGenome g) {
		super(g);
		this.objectiveNum = g.objectiveNum;
		objectiveValue = new double[objectiveNum];
		for (int i = 0; i < objectiveNum; i++) {
			objectiveValue[i] = g.objectiveValue[i];
		}
	}

	public double getFitnessValue() {
		return rank;
	}

	public void calculateObjectiveValue(int testNumber) {
		if (!isEvaluated) {
			for (int i = 0; i < objectiveValue.length; i++) {
				if (GeneticAlgorithm.USE_PREDEFINED_FUNCTION) {
					objectiveValue[i] = MultiObjectiveTestFunction
							.calculateFunctionValue(
									MultiObjectiveTestFunction.FUNCTION_LIST[testNumber][i],
									this.chromosome);
				} else {
					objectiveValue[i] = ((MultiObjectiveRealCodedGeneticAlgorithm) ga).expr[i]
							.compute(this.chromosome[0], this.chromosome[1],
									this.chromosome[2]);
				}
			}
			isEvaluated = true;
			ga.evaluationCount++;
		}
	}

	public boolean dominate(MultiObjectiveRealCodedGenome g, boolean maximize) {
		boolean equal = true;
		for (int i = 0; i < objectiveValue.length; i++) {
			if ((maximize && g.objectiveValue[i] > this.objectiveValue[i])
					|| (!maximize && this.objectiveValue[i] > g.objectiveValue[i]))
				return false;
			else if (this.objectiveValue[i] != g.objectiveValue[i])
				equal = false;
		}
		if (equal)
			return false;
		else
			return true;
	}
}

class PopulationFactory {

	public static <T> ArrayList<Genome<T>> get1DPopulation(Genome<T> g,
			int populationSize) {
		ArrayList<Genome<T>> population = new ArrayList<Genome<T>>();
		for (int i = 0; i < populationSize; i++) {
			population.add(g.randomClone());
		}
		return population;
	}

	public static <T> ArrayList<Genome<T>> get1DPopulation(
			ArrayList<ArrayList<Genome<T>>> population2D) {
		ArrayList<Genome<T>> population = new ArrayList<Genome<T>>();
		for (int i = 0; i < population2D.size(); i++) {
			population.addAll(population2D.get(i));
		}
		return population;
	}

	public static <T> ArrayList<ArrayList<Genome<T>>> get2DPopulation(
			Genome<T> g, int populationSizeX, int populationSizeY) {
		ArrayList<ArrayList<Genome<T>>> population = new ArrayList<ArrayList<Genome<T>>>();
		for (int i = 0; i < populationSizeX; i++) {
			ArrayList<Genome<T>> tempPopulation = new ArrayList<Genome<T>>();
			for (int j = 0; j < populationSizeY; j++) {
				tempPopulation.add(g.randomClone());
			}
			population.add(tempPopulation);
		}
		return population;
	}

	public static <T> ArrayList<ArrayList<Genome<T>>> get2DPopulation(
			ArrayList<Genome<T>> population1D, int populationSizeX) {
		ArrayList<ArrayList<Genome<T>>> population = new ArrayList<ArrayList<Genome<T>>>();
		int populationSizeY = population1D.size() / populationSizeX;
		for (int i = 0; i < populationSizeX; i++) {
			ArrayList<Genome<T>> tempPopulation = new ArrayList<Genome<T>>();
			tempPopulation.addAll(population1D.subList(populationSizeY * i,
					populationSizeY * (i + 1)));
			population.add(tempPopulation);
		}
		return population;
	}

	public static <T> ArrayList<ArrayList<Genome<T>>> get2DPopulationClone(
			ArrayList<ArrayList<Genome<T>>> population2D) {
		ArrayList<ArrayList<Genome<T>>> population = new ArrayList<ArrayList<Genome<T>>>();
		for (int i = 0; i < population2D.size(); i++) {
			ArrayList<Genome<T>> temp = new ArrayList<Genome<T>>();
			temp.addAll(population2D.get(i));
			population.add(temp);
		}
		return population;
	}

	public static <T> ArrayList<Genome<T>> get1DNeighborhoodPopulation(
			ArrayList<Genome<T>> population, int activeIndex,
			int[] selectionPattern) {
		ArrayList<Genome<T>> selectedPopulation = new ArrayList<Genome<T>>();
		for (int i : selectionPattern) {
			selectedPopulation.add(population.get((activeIndex + i + population
					.size())
					% population.size()));
		}
		return selectedPopulation;
	}

	public static <T> ArrayList<ArrayList<Genome<T>>> get2DNeighborhoodPopulation(
			ArrayList<ArrayList<Genome<T>>> population, int activeIndexX,
			int activeIndexY, int neighborhoodSize) {
		int neighborhoodSizeX = neighborhoodSize, neighborhoodSizeY = neighborhoodSize;
		if (neighborhoodSize > (population.size() - 1) / 2)
			neighborhoodSizeX = (population.size() - 1) / 2;
		if (neighborhoodSize > (population.get(0).size() - 1) / 2)
			neighborhoodSizeY = (population.get(0).size() - 1) / 2;
		ArrayList<ArrayList<Genome<T>>> selectedPopulation = new ArrayList<ArrayList<Genome<T>>>();
		int x, y;
		for (int i = -neighborhoodSizeX; i <= neighborhoodSizeX; i++) {
			ArrayList<Genome<T>> temp = new ArrayList<Genome<T>>();
			for (int j = -neighborhoodSizeY; j <= neighborhoodSizeY; j++) {
				x = (activeIndexX + i + population.size()) % population.size();
				y = (activeIndexY + j + population.get(0).size())
						% population.get(0).size();
				temp.add(population.get(x).get(y));
			}
			selectedPopulation.add(temp);
		}
		return selectedPopulation;
	}

	public static <T> boolean isNeighborhoodEquals(
			ArrayList<Genome<T>> neighborhood1,
			ArrayList<Genome<T>> neighborhood2) {
		HashSet<Genome<T>> neighborhoodSet1 = new HashSet<Genome<T>>(), neighborhoodSet2 = new HashSet<Genome<T>>();
		neighborhoodSet1.addAll(neighborhood1);
		neighborhoodSet2.addAll(neighborhood2);
		return neighborhoodSet1.equals(neighborhoodSet2);
	}
}
